/*
 * Copyright (c) 2010 Sendmail, Inc. and its suppliers.
 *      All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: dkim.c,v 1.14 2013/12/03 02:52:22 ca Exp $")
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/memops.h"
#include "sm/time.h"
#include "sm/heap.h"
#include "sm/io.h"
#include "sm/dkimdef.h"
#include "sm/dkim.h"
#include "sm/sha.h"
#include "dkim.h"

#if MTA_USE_DKIM
static open_F		sm_dkimopen;
static write_F		sm_dkimwrite;
static close_F		sm_dkimclose;
static getinfo_F	sm_dkimgetinfo;
static setinfo_F	sm_dkimsetinfo;

sm_stream_T SmDKIMSt =
	SM_STREAM_STRUCT(sm_dkimopen, sm_dkimclose, NULL, sm_dkimwrite,
		NULL, NULL, NULL, sm_dkimgetinfo, sm_dkimsetinfo);
#if DKIM_DEBUG
/* extern int Verbose; */
#endif

/*
**  DKIM_COOKIE_FREE -- free a DKIM cookie
**
**	Parameters:
**		dkim_cookie -- DKIM cookie
**
**	Returns:
**		none
**
**	Last code review:
**	Last code change:
*/

static void
dkim_cookie_free(dkim_cookie_P dkim_cookie)
{
	if (dkim_cookie == NULL)
		return;
	sm_free_size(dkim_cookie, sizeof(*dkim_cookie));
	return;
}

/*
**  DKIM_COOKIE_NEW -- create a DKIM cookie
**
**	Parameters:
**		dkim_msg -- DKIM msg context
**
**	Returns:
**		DKIM cookie (NULL on error [ENOMEM])
**
**	Last code review:
**	Last code change:
*/

static dkim_cookie_P
dkim_cookie_new(dkim_msg_P dkim_msg)
{
	dkim_cookie_P dkim_cookie;

	dkim_cookie = (dkim_cookie_P) sm_zalloc(sizeof(*dkim_cookie));
	if (dkim_cookie == NULL)
		return NULL;
	dkim_cookie->dkf_msg = dkim_msg;
	return dkim_cookie;
}

/*
**  SM_DKIMOPEN -- open a file for DKIM
**
**	Parameters:
**		fp -- file pointer to be associated with the open
**		info -- unused
**		flags -- flags
**		ap -- further arguments
**
**	Returns:
**		usual sm_error code
**
**	Last code review:
**	Last code change:
*/

sm_ret_T
sm_dkimopen(sm_file_T *fp, const void *info, int flags, va_list ap)
{
	int k, h;
	dkim_cookie_P dkim_cookie;
	dkim_msg_P dkim_msg;

	dkim_msg = (dkim_msg_P) info;
	dkim_cookie = dkim_cookie_new(dkim_msg);
	if (dkim_cookie == NULL)
		return sm_error_temp(SM_EM_DKIM, ENOMEM);
	f_cookie(*fp) = (void *) dkim_cookie;

	/* set up DKIM signing ... */
	for (;;) {
		k = va_arg(ap, uint);
		if (SM_IO_WHAT_END == k)
			break;

		switch (k) {
		  case SM_IO_WHAT_HASH:
			h = va_arg(ap, int);
			dkim_cookie->dkf_sha256 = (DKIM_SM_RSA_SHA256 == h);
			break;
		  default:
			/* silently ignore bogus options? */
			break;
		}
	}

	if (dkim_cookie->dkf_sha256)
		SHA256_Init(&DKIM_SHA256_CTX(dkim_cookie));
	else
		SHA1_Init(&DKIM_SHA1_CTX(dkim_cookie));

	return SM_SUCCESS;
}

/*
**  SM_DKIMWRITE -- write to DKIM
**
**	Parameters:
**		fp -- file pointer ro write to
**		buf -- location of data to be written
**		n -- number of bytes to write
**		byteswritten -- number of bytes written (output)
**
**	Returns:
**		usual sm_error code
**
**	Last code review:
**	Last code change:
*/

static sm_ret_T
sm_dkimwrite(sm_file_T *fp, const uchar *buf, size_t n, ssize_t *byteswritten)
{
	dkim_cookie_P dkim_cookie;

	SM_REQUIRE(fp != NULL);
	SM_REQUIRE(f_cookie(*fp) != NULL);
	dkim_cookie = (dkim_cookie_P)f_cookie(*fp);
	if (dkim_cookie->dkf_sha256)
		SHA256_Update(&DKIM_SHA256_CTX(dkim_cookie), buf, n);
	else
		SHA1_Update(&DKIM_SHA1_CTX(dkim_cookie), buf, n);
#if DKIM_DEBUG
	sm_io_fprintf(smioerr, "dkimwrite: buf=\"%.*s\"\n", (int) n, buf);
#endif
	*byteswritten = n;
	return SM_SUCCESS;
}

/*
**  SM_DKIMCLOSE -- close DKIM file
**
**	Parameters:
**		fp -- the file pointer to close
**		flags -- flags
**
**	Returns:
**		usual sm_error code
**
**	Last code review:
**	Last code change:
*/

static sm_ret_T
sm_dkimclose(sm_file_T *fp, int flags)
{
	sm_ret_T ret;
	dkim_cookie_P dkim_cookie;

	SM_REQUIRE(fp != NULL);
	ret = SM_SUCCESS;

	/* already closed? (due to abort) */
	if (f_cookie(*fp) == NULL)
		return ret;
	dkim_cookie = (dkim_cookie_P)f_cookie(*fp);
	dkim_cookie_free((dkim_cookie_P) f_cookie(*fp));
	f_cookie(*fp) = NULL;
	return ret;
}

/*
**  SM_DKIMSETINFO -- set/modify information for a file
**
**	Parameters:
**		fp -- file to set info for
**		what -- type of info to set
**		valp -- location of data used for setting
**
**	Returns:
**		usual sm_error code.
**
**	Last code review:
**	Last code change:
*/

/* ARGSUSED2 */
static sm_ret_T
sm_dkimsetinfo(sm_file_T *fp, int what, void *valp)
{
	switch (what) {
	  default:
		return sm_error_perm(SM_EM_DKIM, EINVAL);
	}
}

/*
**  SM_DKIMGETINFO -- get information about the open file
**
**	Parameters:
**		fp -- file to get info for
**		what -- type of info to get
**		valp -- location to place for info
**
**	Returns:
**		Success: may or may not place info in 'valp' depending
**			on 'what' value, and returns values >=0. Return
**			value may be the obtained info
**		Failure: usual sm_error code.
**
**	Last code review:
**	Last code change:
*/

static sm_ret_T
sm_dkimgetinfo(sm_file_T *fp, int what, void *valp)
{
	dkim_cookie_P dkim_cookie;

	SM_REQUIRE(fp != NULL);

	switch (what) {
	  case SM_IO_WHAT_HASH:
		/* already closed? (due to abort) */
		if (f_cookie(*fp) == NULL)
			return sm_error_perm(SM_EM_DKIM, EINVAL);
		SM_REQUIRE(valp != NULL);
		dkim_cookie = (dkim_cookie_P)f_cookie(*fp);
		if (dkim_cookie->dkf_sha256) {
			SHA256_Final(DKIM_SHA256_OUT(dkim_cookie),
				&DKIM_SHA256_CTX(dkim_cookie));
			*(uchar **)valp = DKIM_SHA256_OUT(dkim_cookie);
		}
		else {
			SHA1_Final(DKIM_SHA1_OUT(dkim_cookie), &DKIM_SHA1_CTX(dkim_cookie));
			*(uchar **)valp = DKIM_SHA1_OUT(dkim_cookie);
		}
		break;

	  default:
		return sm_error_perm(SM_EM_DKIM, EINVAL);
	}
	return SM_SUCCESS;
}
#endif /* MTA_USE_DKIM */
