/*
 * Copyright (c) 2011 Claus Assmann
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the license/LICENSE.3C file which can be found at the
 * top level of this source code distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: ht.c,v 1.8 2011/03/01 22:31:15 ca Exp $")

#include "sm/assert.h"
#include "sm/error.h"
#include "sm/ctype.h"
#include "sm/memops.h"
#include "sm/heap.h"
#include "sm/queue.h"
#include "sm/mta.h"
#include "sm/edbc.h"
#include "sm/base64.h"
#include "sm/io.h"
#include "edbc.h"

static int sm_b64digs2val[] = SM_B64DIGS2VAL;

/*	Note: locking must be provided by caller. */

/* Check whether two keys x and y are equal */
#define HT_KEY_EQ(x, y) ((x)[0] == (y)[0] && sm_memeq((x), (y), SMTP_RCPTID_LEN))

static uint
b16_uint(const char *key, uint b, uint e)
{
	ulong h;
	uint off;
	int c;

	for (off = b, h = 0; off <= e; off++) {
		c = (uchar)key[off];
		if (isascii(c) && isdigit(c))
			c -= '0';
		else if (isascii(c) && isalpha(c))
			c -= isupper(c) ? 'A' - 10 : 'a' - 10;
		h <<= 4;
		h += c;
	}
	return h;
}

static uint
b64_uint(const char *key, uint b, uint e)
{
	ulong h;
	uint off;
	int c, b64;

	for (off = b, h = 0; off <= e; off++) {
		c = (uchar)key[off];
		if (isascii(c) && isdigit(c))
			c -= '0';
		else if (c >= SM_B64MIN && c <= SM_B64MAX &&
			 (b64 = sm_b64digs2val[c - SM_B64MIN]) != -1)
			c = b64;
		h <<= 6;
		h += c;
	}
	return h;
}

/*
**  EDBC_HASH -- hash a key
**
**	Parameters:
**		key -- key to hash
**
**	Returns:
**		hash value
**
**	Last code review:
**	Last code change:
*/

uint
edbc_hash(const char *key)
{
	ulong h;

	h = b16_uint(key, 9, 16);
	h += b64_uint(key, SMTP_RCPTID_LEN - 4, SMTP_RCPTID_LEN - 1);
	return h;
}

/*
**  EDBC_HT_LINK -- insert edbc_node into edbc_ctx (prepend it)
**
**	Parameters:
**		edbc_ctx -- hash edbc_ctx
**		edbc_node -- element
**
**	Returns:
**		It's a macro...
**
**	Last code review:
**	Last code change:
*/

#define EDBC_HT_LINK(edbc_ctx, edbc_node) do {				\
		edbc_head_P ehd = edbc_ctx->edbc_data +		\
				edbc_hash(edbc_node->ecn_rcpt_id) % edbc_ctx->edbc_size;\
		HTINCRACOL(edbc_ctx, ehd); \
		LIST_INSERT_HEAD(ehd, edbc_node, ecn_htl);	\
		HTCHAINLEN(edbc_ctx, ehd, edbc_node);	\
		edbc_ctx->edbc_used++;				\
	} while (0)

/*
**  EDBC_HT_GROW -- extend existing edbc_ctx ht (double its size)
**
**	Parameters:
**		edbc_ctx -- (return) pointer to edbc_ctx
**
**	Returns:
**		usual sm_error code; ENOMEM
**
**	Last code review:
**	Last code change:
*/

static sm_ret_T
edbc_ht_grow(edbc_ctx_P edbc_ctx)
{
	uint old_size, i;
	edbc_node_P edbc_node, edbc_node_next;
	edbc_head_P hd, old_hd;
	sm_ret_T ret;

	SM_IS_EDBC(edbc_ctx);
	old_size = edbc_ctx->edbc_size;
	hd = edbc_ctx->edbc_data;
	old_hd = hd;
	ret = edbc_ht_init(edbc_ctx, 2 * old_size);
	if (sm_is_err(ret))
		return ret;

	/* copy old entries into new edbc_ctx */
	for (i = old_size; i > 0; i--) {
		for (edbc_node = LIST_FIRST(hd);
			edbc_node!= LIST_END(hd);
			edbc_node = edbc_node_next)
		{
			SM_IS_EDBN(edbc_node);
	    	edbc_node_next = LIST_NEXT(edbc_node, ecn_htl);
			EDBC_HT_LINK(edbc_ctx, edbc_node);
		}
		++hd;
	}

	sm_free((void *) old_hd);
	return SM_SUCCESS;
}

/*
**  EDBC_HT_ADD -- add edbc_node
**	Note: this does not check whether key exists! It will simply add
**		the new pair.
**
**	Parameters:
**		edbc_ctx -- edbc_ctx
**		edbc_node -- edbc_node
**
**	Returns:
**		usual sm_error code; ENOMEM, SM_E_FULL
**
**	Side Effects: may grow edbc_ctx.
**
**	Last code review:
**	Last code change:
*/

sm_ret_T
edbc_ht_add(edbc_ctx_P edbc_ctx, edbc_node_P edbc_node)
{
	sm_ret_T ret;

	SM_IS_EDBC(edbc_ctx);
	SM_IS_EDBN(edbc_node);

	if (edbc_ctx->edbc_limit > 0 && edbc_ctx->edbc_used >= edbc_ctx->edbc_limit)
		return sm_error_temp(SM_EM_HT, SM_E_FULL);
	if (edbc_ctx->edbc_used >= edbc_ctx->edbc_size) {
		ret = edbc_ht_grow(edbc_ctx);
		if (sm_is_err(ret))
			return ret;
	}
	EDBC_HT_LINK(edbc_ctx, edbc_node);
	htincradd(edbc_ctx);
	return SM_SUCCESS;
}

/*
**  EDBC_HT_FIND -- lookup value
**
**	Parameters:
**		edbc_ctx -- pointer to edbc_ctx
**		key -- key to hash
**
**	Returns:
**		pointer to value
**
**	Last code review:
**	Last code change:
*/

edbc_node_P
edbc_ht_find(edbc_ctx_P edbc_ctx, rcpt_id_T rcpt_id)
{
	edbc_node_P edbc_node;
	edbc_head_P hd;

	SM_IS_EDBC(edbc_ctx);
	hd = edbc_ctx->edbc_data + edbc_hash(rcpt_id) % edbc_ctx->edbc_size;
	htincrlu(edbc_ctx);
	htinitchain(edbc_ctx);
	LIST_FOREACH(edbc_node, hd, ecn_htl) {
		SM_IS_EDBN(edbc_node);
		if (HT_KEY_EQ(rcpt_id, edbc_node->ecn_rcpt_id))
			return edbc_node;
		htincrlcol(edbc_ctx);
		HTINCRCHAIN(edbc_ctx);
	}
	return NULL;
}

/*
**  EDBC_HT_RM -- remove first entry that matches rcpt_id
**
**	Parameters:
**		edbc_ctx -- edbc_ctx
**		rcpt_id -- rcpt_id to look up
**
**	Returns:
**		usual
**
**	Last code review:
**	Last code change:
*/

sm_ret_T
edbc_ht_rm(edbc_ctx_P edbc_ctx, rcpt_id_T rcpt_id)
{
	edbc_head_P hd;
	edbc_node_P edbc_node, edbc_node_next;

	SM_IS_EDBC(edbc_ctx);
	hd = edbc_ctx->edbc_data + edbc_hash(rcpt_id) % edbc_ctx->edbc_size;
	for (edbc_node = LIST_FIRST(hd);
		edbc_node != LIST_END(hd);
		edbc_node = edbc_node_next)
	{
		SM_IS_EDBN(edbc_node);
	    edbc_node_next = LIST_NEXT(edbc_node, ecn_htl);
		if (HT_KEY_EQ(rcpt_id, edbc_node->ecn_rcpt_id)) {
			LIST_REMOVE(edbc_node, ecn_htl);
			SM_ASSERT(edbc_ctx->edbc_used > 0);
			edbc_ctx->edbc_used--;
			return SM_SUCCESS;
		}
	}
	return SM_E_NOTFOUND;
}

void
edbc_ht_stats(edbc_ctx_P edbc_ctx, char *out, uint len)
{
	SM_IS_EDBC(edbc_ctx);
	SM_REQUIRE(out != NULL);
	SM_REQUIRE(len > 0);
#if SM_HT_PERF
	sm_snprintf(out, len, "adds              %9lu\n"
		"lookups           %9lu\n"
		"deletes           %9lu\n"
		"add collisions    %9lu\n"
		"lookup collisions %9lu\n"
		"delete collisions %9lu\n"
		"max chain len     %9u\n"
		"grows             %9u\n"
		, edbc_ctx->edbc_adds, edbc_ctx->edbc_lookups, edbc_ctx->edbc_deletes
		, edbc_ctx->edbc_acollisions
		, edbc_ctx->edbc_lcollisions, edbc_ctx->edbc_dcollisions
		, edbc_ctx->edbc_chainmax, edbc_ctx->edbc_grows);
#else /* SM_HT_PERF */
	*out = '\0';
#endif /* SM_HT_PERF */
}

#if EDBC_HT_DEBUG
sm_ret_T
edbc_ht_chk(edbc_ctx_P edbc_ctx)
{
	uint i, j;
	edbc_node_P edbc_node;
	edbc_head_P hd;

	SM_IS_EDBC(edbc_ctx);
	hd = edbc_ctx->edbc_data;
	j = 0;
	for (i = edbc_ctx->edbc_size; i > 0; i--) {
		for (j = 0, edbc_node = LIST_FIRST(hd);
			edbc_node!= LIST_END(hd);
			edbc_node = LIST_NEXT(edbc_node, ecn_htl), j++)
		{
			edbc_tnode_P edbc_tnode;
			edbc_node_P edbc_node_nxt;

			SM_IS_EDBN(edbc_node);
			edbc_tnode = edbc_node->ecn_tnode;
			SM_ASSERT(edbc_tnode != NULL);
			edbc_node_nxt = ECNL_FIRST(&edbc_tnode->ectn_hd);
			SM_IS_EDBN(edbc_node_nxt);
		}
		++hd;
	}
	return j + i;
}
#endif /* EDBC_HT_DEBUG */
