/*
 * Copyright (c) 2004 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 *
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: smtlsi.c,v 1.19 2014/03/04 05:38:38 ca Exp $")
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/io.h"
#include "sm/tls.h"

#if MTA_USE_TLS

/*
**  TLSI_NEW -- allocate new TLS information context
**
**	Parameters:
**		tls_ctx -- TLS library context
**		ptlsi_ctx -- (pointer to) TLS information context (output)
**
**	Returns:
**		usual error code
*/

sm_ret_T
tlsi_new(tlsi_ctx_P *ptlsi_ctx, tls_ctx_P tls_ctx)
{
	tlsi_ctx_P tlsi_ctx;

	SM_REQUIRE(ptlsi_ctx != NULL);
	*ptlsi_ctx = NULL;
	tlsi_ctx = (tlsi_ctx_P) sm_zalloc(sizeof(*tlsi_ctx));
	if (tlsi_ctx == NULL)
		return sm_error_temp(SM_EM_TLS, ENOMEM);
	*ptlsi_ctx = tlsi_ctx;

	tlsi_ctx->tlsi_vrfy = TLS_VRFY_NONE;
	tlsi_ctx->tlsi_cipher_bits = -1;
	tlsi_ctx->tlsi_algs_bits = -1;
	tlsi_ctx->tlsi_tls = tls_ctx;
	tlsi_ctx->tlsi_log_field = NULL;
	tlsi_ctx->tlsi_log_value = NULL;

	/* allocate strings here? */

	return SM_SUCCESS;
}

/*
**  TLSI_FREE -- free TLS information context
**
**	Parameters:
**		tlsi_ctx -- TLS information context
**
**	Returns:
**		usual error code (SUCCESS)
*/

sm_ret_T
tlsi_free(tlsi_ctx_P tlsi_ctx)
{
	int i;

	if (NULL == tlsi_ctx)
		return SM_SUCCESS;
	SM_STR_FREE(tlsi_ctx->tlsi_cipher);
	SM_STR_FREE(tlsi_ctx->tlsi_version);
	SM_STR_FREE(tlsi_ctx->tlsi_cert_subject);
	SM_STR_FREE(tlsi_ctx->tlsi_cert_issuer);
	SM_STR_FREE(tlsi_ctx->tlsi_cn_subject);
	SM_STR_FREE(tlsi_ctx->tlsi_cn_issuer);
	SM_STR_FREE(tlsi_ctx->tlsi_vrfy_err);
	for (i = 0; i < SM_ARRAY_SIZE(tlsi_ctx->tlsi_altnames); i++)
		SM_STR_FREE(tlsi_ctx->tlsi_altnames[i]);
	for (i = 0; i < SM_ARRAY_SIZE(tlsi_ctx->tlsi_ca_subjects); i++)
		SM_STR_FREE(tlsi_ctx->tlsi_ca_subjects[i]);
	SM_STR_FREE(tlsi_ctx->tlsi_cert_fp);
#if MTA_TLS_NOTAFTER
	SM_STR_FREE(tlsi_ctx->tlsi_notAfter);
#endif
	tlsi_ctx->tlsi_tls = NULL;
	sm_free_size(tlsi_ctx, sizeof(*tlsi_ctx));
	return SM_SUCCESS;
}

/*
**  TLSI_CLR -- clear TLS information context
**
**	Parameters:
**		tlsi_ctx -- TLS information context
**
**	Returns:
**		usual error code (SUCCESS)
*/

sm_ret_T
tlsi_clr(tlsi_ctx_P tlsi_ctx)
{
	int i;

	if (NULL == tlsi_ctx)
		return SM_SUCCESS;

	tlsi_ctx->tlsi_vrfy = TLS_VRFY_NONE;
	tlsi_ctx->tlsi_cipher_bits = -1;
	tlsi_ctx->tlsi_algs_bits = -1;

	SM_STR_CLR(tlsi_ctx->tlsi_cipher);
	SM_STR_CLR(tlsi_ctx->tlsi_version);
	SM_STR_CLR(tlsi_ctx->tlsi_cert_subject);
	SM_STR_CLR(tlsi_ctx->tlsi_cert_issuer);
	SM_STR_CLR(tlsi_ctx->tlsi_cn_subject);
	SM_STR_CLR(tlsi_ctx->tlsi_cn_issuer);
	SM_STR_CLR(tlsi_ctx->tlsi_vrfy_err);
	for (i = 0; i < SM_ARRAY_SIZE(tlsi_ctx->tlsi_altnames); i++)
		SM_STR_CLR(tlsi_ctx->tlsi_altnames[i]);
	for (i = 0; i < SM_ARRAY_SIZE(tlsi_ctx->tlsi_ca_subjects); i++)
		SM_STR_CLR(tlsi_ctx->tlsi_ca_subjects[i]);
	SM_STR_CLR(tlsi_ctx->tlsi_cert_fp);
#if MTA_TLS_NOTAFTER
	SM_STR_CLR(tlsi_ctx->tlsi_notAfter);
#endif
#if FFR_CERT_PINNING
	sm_memzero(tlsi_ctx->tlsi_cert_md, sizeof(tlsi_ctx->tlsi_cert_md));
	tlsi_ctx->tlsi_cert_mdlen = 0;
	tlsi_ctx->tlsi_hs_stat = 0;
#endif
	tlsi_ctx->tlsi_log_field = NULL;
	tlsi_ctx->tlsi_log_value = NULL;
	return SM_SUCCESS;
}

#endif /* MTA_USE_TLS */
