/*
 *  consumer program emulating smtps receiver side in qmgr.
 *  uses pthreads directly.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-consq-0.c,v 1.6 2013/12/08 16:35:52 ca Exp $")

#include "sm/assert.h"
#include "sm/error.h"
#include "sm/memops.h"
#include "sm/heap.h"
/* #include "sm/test.h" */
#include "sm/fxszq.h"
#include "sm/socket.h"
#include "sm/unixsock.h"
#include "sm/io.h"
#include "sm/misc.h"

#include <pthread.h>

static pthread_mutex_t	 optas_mutex, iq_mutex, iqf_mutex, iqdb_mutex;
static pthread_cond_t	 optas_cond, iq_cond;
static int iqbd_max, iqdb_cur;
/* static bool iqdb_notified = false; */
static pthread_t task1, task2, iqdb_task;
static int fd;
static long optas_add, optas_rm;
static long us;

#define IQDB_MAX 1000

#define LOCK_OPTAS r = pthread_mutex_lock(&optas_mutex)
#define UNLOCK_OPTAS r = pthread_mutex_unlock(&optas_mutex)

static sm_ret_T qm_listen(const char *_sockname, int *_pfd);

/*
need two queues:
1. "open transactions" (optas, fxszq)
2. incoming queue (iqdb, queue)
receiver put entries into optas and iqdb.
"ctaid" is woken up whenever optas is full (and after a certain amount of time)
and removes all entries from it and sends "OKs" back.
another thread is taking out entries of iqdb at some rate.
we can simulate this by using a counter, there's no need for a queue.
*/

struct fxszq_S
{
	uint	 fsq_max;   /* allocated size */
	uint	 fsq_cur;   /* currently used (basically last-first) */
	uint	 fsq_first; /* first index to read */
	uint	 fsq_last;  /* last index to read (first to write) */
	int	*fsq_array; /* array */
};
typedef struct fxszq_S	fxszq_T, *fxszq_P;
fxszq_T optas;

#include "t-prdcns-0.h"

struct t_ctx_S
{
	int			 t_called;
	int			 t_once;
};
typedef struct t_ctx_S t_ctx_T, *t_ctx_P;

/* initialize structures, invoke listener */
static int
t_init(uint sz)
{
	int r;
	sm_ret_T ret;

	SM_TEST(sz != 0);
	if (sz == 0)
		return -1;
	optas.fsq_array = sm_malloc(sizeof(int) * sz);
	SM_TEST(optas.fsq_array != NULL);
	if (optas.fsq_array == NULL)
		return -1;
	FSQ_INIT(sz, optas.fsq_max, optas.fsq_cur, optas.fsq_first, optas.fsq_last);
	SM_TEST(FSQ_IS_EMPTY(optas.fsq_cur));
	SM_TEST(!FSQ_IS_FULL(optas.fsq_max, optas.fsq_cur));
	r = pthread_mutex_init(&optas_mutex, NULL);
	SM_TEST(r == 0);
	r = pthread_cond_init(&optas_cond, NULL);
	SM_TEST(r == 0);

	r = pthread_mutex_init(&iq_mutex, NULL);
	SM_TEST(r == 0);
	r = pthread_cond_init(&iq_cond, NULL);
	SM_TEST(r == 0);
	r = pthread_mutex_init(&iqdb_mutex, NULL);
	SM_TEST(r == 0);
	r = pthread_mutex_init(&iqf_mutex, NULL);
	SM_TEST(r == 0);

	ret = qm_listen(MYSOCK, &fd);
	return ret;
}

/* add an item i to iqdb, return #free entries */
static int
add_iqdb(int i)
{
	int r, n;

	r = pthread_mutex_lock(&iqdb_mutex);
	SM_TEST(0 == r);
	if (iqbd_max <= iqdb_cur) {
		r = pthread_mutex_unlock(&iqdb_mutex);
		SM_TEST(0 == r);
		return 0;
	}
	++iqdb_cur;
	n = iqbd_max - iqdb_cur;
	r = pthread_mutex_unlock(&iqdb_mutex);
	SM_TEST(0 == r);
	return n;
}

/* remove one item from iqdb, return #free entries */
static int
rm_iqdb(void)
{
	int r, n;

	r = pthread_mutex_lock(&iqdb_mutex);
	SM_TEST(0 == r);
	if (iqdb_cur == 0) {
		r = pthread_mutex_unlock(&iqdb_mutex);
		SM_TEST(0 == r);
		return -1;
	}
	--iqdb_cur;
	n = iqdb_cur;
	r = pthread_mutex_unlock(&iqdb_mutex);
	SM_TEST(0 == r);
	return n;
}

/* return #free entries */
static int
nfree_iqdb(void)
{
	int r, n;

	r = pthread_mutex_lock(&iqdb_mutex);
	SM_TEST(0 == r);
	n = iqbd_max - iqdb_cur;
	r = pthread_mutex_unlock(&iqdb_mutex);
	SM_TEST(0 == r);
	return n;
}

/* add an item i to optas, return percentage free entries */
static int
add_optas(int i)
{
	int r, p;

	LOCK_OPTAS;
	if (FSQ_IS_FULL(optas.fsq_max, optas.fsq_cur)) {
		UNLOCK_OPTAS;
		return 100;
	}
	FSQ_APPEND(optas.fsq_max, optas.fsq_last, optas.fsq_cur, optas.fsq_array, i);
	SM_TEST(!FSQ_IS_EMPTY(optas.fsq_cur));
	++optas_add;
	if (optas_add > reps) stop = true;
	p = FSQ_PERC(optas.fsq_max, optas.fsq_cur);
	UNLOCK_OPTAS;
	return p;
}

/*
 * remove *pnelems items from optas, return percentage free entries
 * set #free, #used, #removed, pv: array of entries removed[0..*pnelems]
 */
static int
rm_optas(int *pfree, int *pused, int *pv, int *pnelems)
{
	int i, h, r, n;

	LOCK_OPTAS;
	if (FSQ_IS_EMPTY(optas.fsq_cur)) {
		UNLOCK_OPTAS;
		return 0;
	}
	n = SM_MIN(optas.fsq_cur, *pnelems);
	for (i = 0; i < n; i++) {
		pv[i] = FSQ_GET(optas.fsq_max, optas.fsq_first, optas.fsq_cur, optas.fsq_array, h);
		++optas_rm;
	}
	*pnelems = i;
	h = FSQ_PERC(optas.fsq_max, optas.fsq_cur);
	*pfree = optas.fsq_max - optas.fsq_cur;
	*pused = optas.fsq_cur;
	UNLOCK_OPTAS;
	return h;
}

/* accept a single connection */
static sm_ret_T
qm_listen(const char *sockname, int *pfd)
{
	int lfd, r;
	sm_ret_T ret;
	struct sockaddr addr;
	sockaddr_len_T addrlen;

	SM_REQUIRE(sockname != NULL);
	SM_REQUIRE(pfd != NULL);
	ret = SM_SUCCESS;
	r = unlink(sockname);
	if (r < 0 && errno != ENOENT) {
		ret = sm_err_perm(errno);
		dbg_print("qm_listen: socket=%s, unlink=%m\n", sockname, ret);
		goto error;
	}
	lfd = unix_server_listen(sockname, 10);
	if (!is_valid_socket(lfd)) {
		ret = sm_err_perm(errno);
		dbg_print("qm_listen: socket=%s, unix_server_listen=%m\n", sockname, lfd);
		goto error;
	}

	addrlen = sizeof(addr);
	if (Verbose > 2) dbg_print("qm_listen: accept\n");
	fd = unix_server_accept(lfd, &addr, &addrlen);
	SM_TEST(fd >= 0);
	if (fd < 0) { ret = sm_err_perm(errno); goto error; }
	if (Verbose > 1) dbg_print("qm_listen: accepted fd=%d\n", fd);
	*pfd = fd;

	/* FALLTHROUGH */
  error:
	return ret;
}

/* send a value */
static int
sndv(int32_t v)
{
	int p;
	char buf[PBUFSZ];

	*((int *)buf) = v;
	p = write(fd, buf, PBUFSZ);
	if (p <= 0) return sm_err_perm(errno);
	++q_sent;
	return p;
}

/* receive a value */
static int
rcvv(int32_t *pv)
{
	int r;
	char buf[PBUFSZ];

	r = read(fd, buf, sizeof(buf));
	if (r <= 0) return sm_err_perm(errno);
	++q_rcvd;
	*pv = *((int *)buf);
	return r;
}

/* remove all entries from iqdb and "do" something with them */
static int
iqdbf(int pt)
{
	int r, l, t;

	r = pthread_mutex_trylock(&iqf_mutex);
	if (r != 0)
		return -1;

	t = l = r = rm_iqdb();
	if (r < 0 || Verbose > 2)
		dbg_print("		iqdbf: pt=%x, rm_iqdb=%d\n", pt, r);
	while (l-- > 0) {
		r = rm_iqdb();
		if (r < 0) {
			dbg_print("		iqdbf: pt=%x, where=loop, rm_iqdb=%d, l=%d\n", pt, r, l);
			break;
		}
		if (Verbose > 2)
			dbg_print("		iqdbf: pt=%x, where=loop, rm_iqdb=%d, l=%d\n", pt, r, l);

		/* simulate doing some work by delaying thread execution */
		r = sm_us_sleep(us);
		if (r < 0 || Verbose > 4)
			dbg_print("		iqdbf: pt=%x, where=loop, us_sleep=%d, l=%d\n", pt, r, l);
	}
	r = pthread_mutex_unlock(&iqf_mutex);
	SM_TEST(0 == r);
	return t;
}

/*
 * receive thread: receive data, add it to iqdb and optas;
 * if #free in idbq is 0: free iqdb
 * if #free in idbq is <=10: signal iq (which will free iqdb?)
 * if optas is more than 25% full, signal
 */
static void *
t_rcv(void *ctx)
{
	t_ctx_P fctx;

	SM_ASSERT(ctx != NULL);
	fctx = (t_ctx_P) ctx;

	while (1) {
		int i, r, p, f;

		r = 0;
		if (Verbose > 0)
			dbg_print("t_rcv: c=%d\n", fctx->t_called);
		fctx->t_called++;
		if (stop) return 0;

		for (i = 0; i < fctx->t_once; i++) {
			int32_t v;

			if (Verbose > 2)
				dbg_print("t_rcv: i=%d, where=before_rcvv\n", i);
			r = rcvv(&v);
			if (Verbose > 2)
				dbg_print("t_rcv: i=%d, rcvv=%m, v=%d\n", i, r, v);
			if (r <= 0)
				goto error;
			f = add_iqdb(v);
			if (Verbose > 2 || f <= 0)
				dbg_print("t_rcv: i=%d, add_iqdb=%d\n", i, f);
			if (f <= 0) {
				r = iqdbf(1);
				if (r != 0 || Verbose > 2)
					dbg_print("t_rcv: i=%d, iqdb=%d\n", i, r);
			}
			else if (f <= 10) {
				r = pthread_cond_signal(&iq_cond);
				if (r != 0 || Verbose > 2)
					dbg_print("t_rcv: i=%d, signal_iqdb=%d\n", i, r);
			}
			p = add_optas(v);
			if (Verbose > 2)
				dbg_print("t_rcv: i=%d, add_optas=%d\n", i, p);
			if (p > 25) { /* maximum percentage? */
				r = pthread_cond_signal(&optas_cond);
				if (r != 0 || Verbose > 2)
					dbg_print("t_rcv: i=%d, p=%d, signal_optas=%d\n", i, p, r);
#if 0
				do {
					p = perc_optas();
					if (p > 75) {
						r = sm_us_sleep(100);
						if (r != 0 || Verbose > 2)
							dbg_print("t_rcv: i=%d, p=%d, sleep=%d\n", i, p, r);
					}
					if (stop) return 0;
				} while (p > 75);
#endif
				if (p > 75)
					break;
			}
		}
		if (Verbose > 0)
			dbg_print("t_rcv: r=%d, stop=%d\n", r, stop);

		if (stop) return 0;
	}

 error:
	return 0;
}

/*
 * acknowlegde thread: wait for optas signal, removes several items
 * and sends 'ack' back to producer for each, as well as final
 * "available slots", as long as #free(optas) > 20 (?)
 */
static void *
t_ack(void *ctx)
{
	int i, r, n;
	int elems[T_QS];
	t_ctx_P fctx;

	SM_ASSERT(ctx != NULL);
	fctx = (t_ctx_P) ctx;

	r = 0;
	while (1) {
		int v, p, free, used;

		if (Verbose > 0)
			dbg_print("	t_ack: c=%d\n", fctx->t_called);
		fctx->t_called++;
		if (stop) return 0;

		i = pthread_mutex_lock(&optas_mutex);
		SM_TEST_E(i == 0);
		if (Verbose > 2)
			dbg_print("	t_ack: where=afterlock, stat=%d, stop=%d\n", i, stop);
		while (!stop) {
			struct timespec w;
			struct timeval now;

			gettimeofday(&now, NULL);
			w.tv_sec = now.tv_sec + 5;
			w.tv_nsec = 0;
			if (Verbose > 2)
				dbg_print("	t_ack: where=beforewait\n", i);
			i = pthread_cond_timedwait(&optas_cond, &optas_mutex, &w);
			if (i != 0) {
				dbg_print("	t_ack: where=afterwait, stat=%d\n", i);
				if (ETIMEDOUT == i)
					goto error;
			}
			SM_TEST_E(i == 0);
			if (Verbose > 0)
				dbg_print("	t_ack: afterwait=%d, empty=%d\n", i,
					FSQ_IS_EMPTY(optas.fsq_cur));
			i = pthread_mutex_unlock(&optas_mutex);
			SM_TEST_E(i == 0);
			if (Verbose > 0)
				dbg_print("	t_ack: afterlock=%d\n", i);
			if (stop) return 0;
			break;
		}

		do {
			/* get several elements from queue */
			n = SM_MIN(fctx->t_once, T_QS);
			if (Verbose > 2)
				dbg_print("	t_ack: n=%d, before=rm_optas\n", n);
			p = rm_optas(&free, &used, elems, &n);
			if (n == 0 && Verbose > 0 || Verbose > 2)
				dbg_print("	t_ack: n=%d, free=%d, used=%d\n", n, free, used);

			for (i = 0; i < n; i++) {
				v = elems[i];
				r = sndv(v);
				if (Verbose > 0)
					dbg_print("	t_ack: i=%d, v=%d, sndv=%d\n", i, v, r);
				if (r <= 0) goto error;
			}

			p = nfree_iqdb();
			v = -p;
			r = sndv(v);
			if (Verbose > 0)
				dbg_print("	t_ack: p=%d, sndv=%d, stop=%d\n", p, r, stop);
			if (r <= 0) goto error;
			if (stop) return 0;
			n = 0;
			r = rm_optas(&free, &used, elems, &n);
			if (Verbose > 0)
				dbg_print("	t_ack: free=%d\n", free);
		} while (free > 20);
	}

 error:
	return 0;
}

/*
 *  iqdb thread: wait for iq signal
 */
static void *
t_iqdb(void *ctx)
{
	t_ctx_P fctx;

	SM_ASSERT(ctx != NULL);
	fctx = (t_ctx_P) ctx;
	for (;;) {
		int r, i;
		struct timespec w;

		i = pthread_mutex_lock(&iq_mutex);
		SM_TEST_E(i == 0);
		if (Verbose > 2)
			dbg_print("		t_iqdb: where=afterlock, stat=%d, stop=%d\n", i, stop);
		while (!stop) {
			struct timeval now;

			gettimeofday(&now, NULL);
			w.tv_sec = now.tv_sec + 1;
			w.tv_nsec = 0;
			if (Verbose > 2)
				dbg_print("		t_iqdb: where=beforewait\n", i);
			i = pthread_cond_timedwait(&iq_cond, &iq_mutex, &w);
			if (i != 0) {
				dbg_print("		t_iqdb: where=afterwait, stat=%d, stop=%d\n", i, stop);
				if (ETIMEDOUT == i) {
					i = pthread_mutex_unlock(&iq_mutex);
					SM_TEST_E(i == 0);
					break;
				}
			}
			SM_TEST_E(i == 0);
			i = pthread_mutex_unlock(&iq_mutex);
			SM_TEST_E(i == 0);
			if (Verbose > 0)
				dbg_print("		t_iqdb: afterlock=%d\n", i);
			if (stop) return 0;
			break;
		}
		r = iqdbf(2);
	}
	return NULL;
}

static void
testev(uint sz, int adds, int rms)
{
	int ret;
	t_ctx_T tctx1, tctx2, tctx_iqdb;

	ret = t_init(sz);
	if (ret != 0)
		return;

	tctx1.t_called = 0;
	tctx1.t_once = adds;
	ret = pthread_create(&task1, NULL, t_rcv, (void *) &tctx1);
	SM_TEST(ret == 0);
	SM_TEST(task1 != NULL);

	tctx2.t_called = 0;
	tctx2.t_once = rms;
	ret = pthread_create(&task2, NULL, t_ack, (void *) &tctx2);
	SM_TEST(ret == 0);
	SM_TEST(task2 != NULL);

	tctx_iqdb.t_called = 0;
	tctx_iqdb.t_once = rms;
	ret = pthread_create(&iqdb_task, NULL, t_iqdb, (void *) &tctx_iqdb);
	SM_TEST(ret == 0);
	SM_TEST(iqdb_task != NULL);

	ret = pthread_join(task1, NULL);
	/* SM_TEST(ret == 0); */
	dbg_print("join1=%d\n", ret);

	ret = pthread_join(task2, NULL);
	/* SM_TEST(ret == 0); */
	dbg_print("join2=%d\n", ret);

	dbg_print("sent=%ld, rcvd=%ld, adds=%ld, rms=%ld\n",
		q_sent, q_rcvd, optas_add, optas_rm);
}

static void
usage(const char *prg)
{
	sm_io_fprintf(smioerr, "%s: simple producer/consumer test\n", prg);
	sm_io_fprintf(smioerr, "usage %s [options]\n", prg);
	sm_io_fprintf(smioerr, "options\n");
	sm_io_fprintf(smioerr, "-a n  add n items on each iteration\n");
	sm_io_fprintf(smioerr, "-d n  remove n items on each iteration\n");
	sm_io_fprintf(smioerr, "-m n  use n for size of iqdb\n");
	sm_io_fprintf(smioerr, "-r n  run n loops\n");
	sm_io_fprintf(smioerr, "-s n  use n for size of optas\n");
	sm_io_fprintf(smioerr, "-V    increase verbosity\n");
	sm_io_fprintf(smioerr, "-w n  sleep n us for each item removed from iqdb\n");
}

int
main(int argc, char *argv[])
{
	int c, adds, rms;
	uint sz;

	reps = 1000;
	sz = 100;
	adds = rms = 10;
	us = 1;
	iqbd_max = IQDB_MAX;
	while ((c = getopt(argc, argv, "a:d:m:r:s:Vw:")) != -1) {
		switch (c) {
		  case 'a':
			adds = atoi(optarg);
			break;
		  case 'd':
			rms = atoi(optarg);
			break;
		  case 'm':
			iqbd_max = atoi(optarg);
			break;
		  case 'r':
			reps = atoi(optarg);
			break;
		  case 's':
			sz = atoi(optarg);
			break;
		  case 'V':
			Verbose++;
			break;
		  case 'w':
			us = atoi(optarg);
			break;
		  default:
			usage(argv[0]);
			return(1);
		}
	}
	testev(sz, adds, rms);
	return 0;
}
